//
//  CoreDatabase.m
//
//  Created by Ingo Kasprzak on 03.08.11.
//  Copyright 2010 Silutions. All rights reserved.
//

#import "CoreDatabase.h"

@interface CoreDatabase ()
@property (strong, readonly) NSPersistentStoreCoordinator *persistentStoreCoordinator;

@property (retain) NSString* databaseName;
@property (retain) NSString* datamodelName;

- (void)releaseDatabase;

@end

@implementation CoreDatabase
@synthesize databaseName, datamodelName;

#pragma mark - Public Methods
- (BOOL)saveChanges {
	NSError *error = nil;
    BOOL result = [self saveChanges:&error];
    if (!result) {
        [self handleUnresolvedError:error];        
    }
    return result;
}

- (BOOL)saveChanges:(NSError **)error {
    BOOL result = YES;
    if (managedObjectContext != nil) {
        if ([self.managedObjectContext hasChanges] && ![self.managedObjectContext save:error]) {
			result = NO;
        } 
    }
    return result;
}

- (void) willInitializeManagedObjectContext {
}
- (void) didInitializeManagedObjectContext {
}


#pragma mark - #pragma mark Initialization

- (void)setupCoreDatabaseInstance {
	// called from every init method
}

- (id) init {
	self = [super init];
	if (self != nil) {
		self.databaseName = SHAREDDATABASE_DEFAULT_DATABASENAME;
		self.datamodelName = SHAREDDATABASE_DEFAULT_DATAMODELNAME;
		[self setupCoreDatabaseInstance];
	}
	return self;
}

- (id)initWithDatabaseName:(NSString*)baseName andDatamodelName:(NSString*)modelName {
	self = [super init];
	if (self != nil) {
		self.databaseName = baseName;
		self.datamodelName = modelName;
		[self setupCoreDatabaseInstance];
	}
	return self;
}

- (id)initWithDatabaseName:(NSString*)baseName {
	self = [super init];
	if (self != nil) {
		self.databaseName = baseName;
		self.datamodelName = baseName;
		[self setupCoreDatabaseInstance];
	}
	return self;
}


#pragma mark - Private Methods

- (void)handleUnresolvedError:(NSError*)error {
	DBLog(@"SharedDatabase: Unresolved error %@, %@", error, [error userInfo]);
	//abort();
}

- (void)releaseDatabase {
	managedObjectContext = nil;
	managedObjectModel = nil;
	persistentStoreCoordinator = nil;
}


#pragma mark - Application's Documents directory

/**
 Returns the path to the application's Documents directory.
 */

- (NSString *)findOrCreateDirectory:(NSSearchPathDirectory)searchPathDirectory
                           inDomain:(NSSearchPathDomainMask)domainMask
                appendPathComponent:(NSString *)appendComponent
                              error:(NSError **)errorOut {
    // Search for the path
    NSArray* paths = NSSearchPathForDirectoriesInDomains(searchPathDirectory, domainMask, YES);
    if ([paths count] == 0) {
        // *** creation and return of error object omitted for space
        return nil;
    }
    
    // Normally only need the first path
    NSString *resolvedPath = [paths objectAtIndex:0];
    
    if (appendComponent) {
        resolvedPath = [resolvedPath
                        stringByAppendingPathComponent:appendComponent];
    }
    
    // Create the path if it doesn't exist
    NSError *error = nil;
    BOOL success = [[NSFileManager defaultManager] createDirectoryAtPath:resolvedPath
                                             withIntermediateDirectories:YES
                                                              attributes:nil
                                                                   error:&error];
    if (!success)  {
        if (errorOut) {
            *errorOut = error;
        }
        return nil;
    }
    
    // If we've made it this far, we have a success
    if (errorOut) {
        *errorOut = nil;
    }
    return resolvedPath;
}



- (NSString *)applicationFilesDirectory {
    NSString *executableName = [[[NSBundle mainBundle] infoDictionary] objectForKey:(NSString*)kCFBundleExecutableKey];
    NSError *error = nil;
    NSString *result = [self findOrCreateDirectory:NSApplicationSupportDirectory
                                          inDomain:NSUserDomainMask
                               appendPathComponent:executableName
                                             error:&error];
    if (error) {
        DBLog(@"Unable to find or create application files directory:\n%@", error);
    }
    return result;
}

- (NSString*)storeFileName {
    NSString* fileName = [self.databaseName stringByAppendingPathExtension:@"sqlite"];
    return [[self applicationFilesDirectory] stringByAppendingPathComponent:fileName];
}

#pragma mark - Fetch Helper

- (NSFetchRequest *)fetchRequestTemplateForName:(NSString *)name {
    return [self.managedObjectModel fetchRequestTemplateForName:name];
}

- (id)fetchObjectForEntityClass:(Class)entityClass
				  withPredicate:(id)stringOrPredicate, ... {
	id result = nil;
	NSPredicate *predicate = nil;
	
	if (stringOrPredicate) {
        if ([stringOrPredicate isKindOfClass:[NSString class]]) {
            va_list variadicArguments;
            va_start(variadicArguments, stringOrPredicate);
			predicate = [NSPredicate predicateWithFormat:stringOrPredicate
											   arguments:variadicArguments];
            va_end(variadicArguments);
        } else {
            predicate = (NSPredicate *)stringOrPredicate;
        }
    }
	NSSet *set = [self fetchObjectsForEntityClass:entityClass
									withPredicate:predicate];
	if ([set count]) {
		result = [set anyObject];
	}
	return result;
}

- (NSSet *)fetchObjectsForEntityClass:(Class)entityClass
						withPredicate:(id)stringOrPredicate, ... {
	
	NSString* newEntityName = NSStringFromClass(entityClass);
	
	NSEntityDescription *entity = [NSEntityDescription entityForName:newEntityName
											  inManagedObjectContext:self.managedObjectContext];
	
	NSFetchRequest *request = [[NSFetchRequest alloc] init];
	[request setEntity:entity];
	
	if (stringOrPredicate) {
        NSPredicate *predicate;
        if ([stringOrPredicate isKindOfClass:[NSString class]]) {
            va_list variadicArguments;
            va_start(variadicArguments, stringOrPredicate);
			predicate = [NSPredicate predicateWithFormat:stringOrPredicate
											   arguments:variadicArguments];
            va_end(variadicArguments);
        } else {
            predicate = (NSPredicate *)stringOrPredicate;
        }
        [request setPredicate:predicate];
    }
	
    NSError *error = nil;
    NSArray *results = [self.managedObjectContext executeFetchRequest:request error:&error];
    if (error != nil) {
        [NSException raise:NSGenericException format:@"%@", [error description]];
    }
    return [NSSet setWithArray:results];
}

- (NSArray *)fetchObjectsForEntityClass:(Class)entityClass
						 sortDescriptor:(id)stringOrDescriptorOrArray
						  withPredicate:(id)stringOrPredicate, ... {
	
	NSString* newEntityName = NSStringFromClass(entityClass);
	
	NSEntityDescription *entity = [NSEntityDescription entityForName:newEntityName
											  inManagedObjectContext:self.managedObjectContext];
	
	NSFetchRequest *request = [[NSFetchRequest alloc] init];
	[request setEntity:entity];
	
	if (stringOrPredicate)
    {
        NSPredicate *predicate;
        if ([stringOrPredicate isKindOfClass:[NSString class]])
        {
            va_list variadicArguments;
            va_start(variadicArguments, stringOrPredicate);
			predicate = [NSPredicate predicateWithFormat:stringOrPredicate
											   arguments:variadicArguments];
            va_end(variadicArguments);
        } else {
            predicate = (NSPredicate *)stringOrPredicate;
        }
        [request setPredicate:predicate];
    }
	
	if ([stringOrDescriptorOrArray isKindOfClass:[NSString class]]) {
		[request setSortDescriptors:@[[NSSortDescriptor sortDescriptorWithKey:stringOrDescriptorOrArray
																	ascending:YES]]];
	} else if ([stringOrDescriptorOrArray isKindOfClass:[NSSortDescriptor class]]){
		[request setSortDescriptors:@[stringOrDescriptorOrArray]];
	} else {
		[request setSortDescriptors:stringOrDescriptorOrArray];
	}
	
    NSError *error = nil;
    NSArray *results = [self.managedObjectContext executeFetchRequest:request error:&error];
    if (error != nil) {
        [NSException raise:NSGenericException format:@"%@", [error description]];
    }
    return results;
}

- (NSUInteger)countForEntityClass:(Class)entityClass
					withPredicate:(id)stringOrPredicate, ... {
	NSString* newEntityName = NSStringFromClass(entityClass);
	
	NSEntityDescription *entity = [NSEntityDescription entityForName:newEntityName
											  inManagedObjectContext:self.managedObjectContext];
	
	NSFetchRequest *request = [[NSFetchRequest alloc] init];
	[request setEntity:entity];
	
	if (stringOrPredicate) {
        NSPredicate *predicate;
        if ([stringOrPredicate isKindOfClass:[NSString class]]) {
            va_list variadicArguments;
            va_start(variadicArguments, stringOrPredicate);
			predicate = [NSPredicate predicateWithFormat:stringOrPredicate
											   arguments:variadicArguments];
            va_end(variadicArguments);
        } else {
            predicate = (NSPredicate *)stringOrPredicate;
        }
        [request setPredicate:predicate];
    }
	
	NSError *error = nil;
	NSUInteger result = [self.managedObjectContext countForFetchRequest:request error:&error];
	if (error != nil) {
		[NSException raise:NSGenericException format:@"%@", [error description]];
	}
	return result;
}



#pragma mark - Core Data stack

- (void)removeDatabase {
    [self releaseDatabase];
    NSString *fullPath = [self storeFileName];
    NSFileManager *manager = [NSFileManager defaultManager];
    if ([manager fileExistsAtPath:fullPath]) {
        [manager removeItemAtPath:fullPath error:nil];
    }
}

/**
 Returns the managed object context for the application.
 If the context doesn't already exist, it is created and bound to the persistent store coordinator for the application.
 */
- (NSManagedObjectContext *) managedObjectContext {
	
    if (managedObjectContext != nil) {
        return managedObjectContext;
    }
	
	[self willInitializeManagedObjectContext];
	
    NSPersistentStoreCoordinator *coordinator = [self persistentStoreCoordinator];
    if ((coordinator != nil) && (managedObjectContext==nil)) {
        managedObjectContext = [[NSManagedObjectContext alloc] init];
        [managedObjectContext setPersistentStoreCoordinator: coordinator];
    }
	
	[self didInitializeManagedObjectContext];

    return managedObjectContext;
}

/**
 Returns the managed object model for the application.
 If the model doesn't already exist, it is created by merging all of the models found in the application bundle.
 */
- (NSManagedObjectModel *)managedObjectModel {
	
    if (managedObjectModel != nil) {
        return managedObjectModel;
    }
	
	NSString* modelPath = [[NSBundle mainBundle] pathForResource:self.datamodelName ofType:@"momd"];
	if (!modelPath) {
		modelPath = [[NSBundle mainBundle] pathForResource:self.datamodelName ofType:@"mom"];
	}
	if (!modelPath) {
		NSDictionary *dict = @{ NSFilePathErrorKey : self.datamodelName};
		NSError* error = [NSError errorWithDomain:NSCocoaErrorDomain code:NSFileNoSuchFileError userInfo:dict];
		[self handleUnresolvedError:error];
	}
	
	NSURL *modelURL =  [NSURL fileURLWithPath:modelPath];
    managedObjectModel = [[NSManagedObjectModel alloc] initWithContentsOfURL:modelURL];    
    return managedObjectModel;
}

/**
 Returns the persistent store coordinator for the application.
 If the coordinator doesn't already exist, it is created and the application's store added to it.
 */
- (NSPersistentStoreCoordinator *)persistentStoreCoordinator {
	
    if (persistentStoreCoordinator != nil) {
        return persistentStoreCoordinator;
    }
	
	NSString *storeFile = [self storeFileName];
    NSURL *storeUrl = [NSURL fileURLWithPath:storeFile];

	NSError *error = nil;
    persistentStoreCoordinator = [[NSPersistentStoreCoordinator alloc] initWithManagedObjectModel:[self managedObjectModel]];
    if (![persistentStoreCoordinator addPersistentStoreWithType:NSSQLiteStoreType configuration:nil URL:storeUrl options:nil error:&error]) {
		[self handleUnresolvedError:error];
    }
	
    return persistentStoreCoordinator;
}

#pragma mark - Memory management

- (void)dealloc {
	[self releaseDatabase];
}

@end
