//
//  MultipeerConnectivitySynchronizationServer.m
//  TicTacToe
//
//  Created by Ingo Kasprzak on 11.10.12.
//  Copyright (c) 2012 Silutions. All rights reserved.
//

#define ServiceTypeKey @"0x0-tictactoe"

// Keys for synchronization dictionary
#define SyncDatabaseSyncCommandKey @"command"
#define SyncDatabaseSyncDataKey @"data"
#define SyncDatabaseSyncDateKey @"date"
#define SyncDatabaseSyncSenderUIDKey @"senderUID"
#define SyncDatabaseSyncDataUIDKey @"dataUID"
#define SyncDatabaseSyncWantsReceiptKey @"wantsReceipt"


#import "MultipeerConnectivitySynchronizationServer.h"
#import "SynchronizationServer.h"
#import "Device+Methods.h"

@interface MultipeerConnectivitySynchronizationServer ()
@property (strong) MCPeerID *peerID;
@property (strong) NSMutableSet *discoveredPeers;
@property (strong) NSMutableDictionary *connectedPeerObjects;
@property (strong) MCSession *mcSession;
@property (strong) MCNearbyServiceAdvertiser *nearbyServiceAdvertiser;
@property (strong) MCNearbyServiceBrowser *nearbyServiceBrowser;
@end

@implementation MultipeerConnectivitySynchronizationServer

#pragma mark - Initialization

- (BOOL)startServer {
    if (!self.mcSession) {
		[super startServer];
		
		self.discoveredPeers = [NSMutableSet setWithCapacity:2];
		self.connectedPeerObjects = [NSMutableDictionary dictionaryWithCapacity:2];

        // Standard MultipeerConnectivity initialization
		if (!self.peerID) {
			self.peerID = [[MCPeerID alloc] initWithDisplayName:self.delegate.thisDevice.uid];
		}
		
		self.mcSession = [[MCSession alloc] initWithPeer:self.peerID
										securityIdentity:nil
									encryptionPreference:MCEncryptionNone];
		
		self.mcSession.delegate = self;
		
		
		// Start Service advertiser
		self.nearbyServiceAdvertiser = [[MCNearbyServiceAdvertiser alloc] initWithPeer:self.peerID
																		 discoveryInfo:nil
																		   serviceType:ServiceTypeKey];
		self.nearbyServiceAdvertiser.delegate = self;
		[self.nearbyServiceAdvertiser startAdvertisingPeer];
		
		// Start Service Browser
		self.nearbyServiceBrowser = [[MCNearbyServiceBrowser alloc] initWithPeer:self.peerID
																	 serviceType:ServiceTypeKey];
		self.nearbyServiceBrowser.delegate = self;
		[self.nearbyServiceBrowser startBrowsingForPeers];
		
		
		DBLog(@"MultipeerConnectivity initialized.");
	}

	return YES;
}

- (BOOL)stopServer {
	if (self.mcSession) {
		DBLog(@"Stopping MultipeerConnectivity");
		[super stopServer];
		
		[self.nearbyServiceBrowser stopBrowsingForPeers];
		self.nearbyServiceBrowser = nil;
		
		[self.nearbyServiceAdvertiser stopAdvertisingPeer];
		self.nearbyServiceAdvertiser = nil;

		[self.mcSession disconnect];
		self.mcSession = nil;
	}
	
	self.discoveredPeers = nil;
	self.connectedPeerObjects = nil;
	
    return YES;
}

- (BOOL)sendData:(NSData*)data toPeers:(NSArray*)peerIDs {
	__block BOOL result = YES;
	
	NSMutableArray *peerObjects = [NSMutableArray arrayWithCapacity:[peerIDs count]];
	[peerIDs enumerateObjectsUsingBlock:^(NSString *obj, NSUInteger idx, BOOL *stop) {
		MCPeerID *peer = [self.connectedPeerObjects objectForKey:obj];
		if (peer) {
			[peerObjects addObject:peer];
		} else {
			*stop = YES;
			result = NO;
		}
	}];
	
	if (peerObjects.count==0) { result = NO; }
	
	if (result) {
		[self.mcSession sendData:data toPeers:peerObjects withMode:MCSessionSendDataReliable error:nil];
	}

    return result;
}

- (BOOL)sendData:(NSData*)data toDevices:(NSArray*)deviceUIDs {
	NSArray *peerIDs = [self peerIdentifiersForDeviceUIDs:deviceUIDs];
	return [self sendData:data toPeers:peerIDs];
}

#pragma mark - Service Advertiser Delegate Methods

- (void)advertiser:(MCNearbyServiceAdvertiser *)advertiser didReceiveInvitationFromPeer:(MCPeerID *)peerID
	   withContext:(NSData *)context invitationHandler:(void (^)(BOOL accept, MCSession *session))invitationHandler {
	
	BOOL discoveredHost = [self.discoveredPeers containsObject:peerID.displayName];
	if (!discoveredHost) {
		[self.discoveredPeers addObject:peerID.displayName];
		DBLog(@"Invited by peer %@", peerID.displayName);
	}
	invitationHandler(!discoveredHost, self.mcSession);
}

#pragma mark - Service Browser Delegate Methods

- (void)browser:(MCNearbyServiceBrowser *)browser foundPeer:(MCPeerID *)peerID withDiscoveryInfo:(NSDictionary *)info {
	if (![self.discoveredPeers containsObject:peerID.displayName]) {
		DBLog(@"Found peer %@", peerID.displayName);
		[self.discoveredPeers addObject:peerID.displayName];
		[self.nearbyServiceBrowser invitePeer:peerID toSession:self.mcSession withContext:nil timeout:0];
	} else {
		DBLog(@"Found already discovered peer %@", peerID.displayName);
	}
}
- (void)browser:(MCNearbyServiceBrowser *)browser lostPeer:(MCPeerID *)peerID {
	DBLog(@"Lost peer %@", peerID.displayName);
	[self.discoveredPeers removeObject:peerID.displayName];
}

#pragma mark - Multipeer Connectivity Delegate Methods

- (void)session:(MCSession *)session didStartReceivingResourceWithName:(NSString *)resourceName
	   fromPeer:(MCPeerID *)peerID withProgress:(NSProgress *)progress {
	DBLog(@"This app is not configured to receive data with method %s", __PRETTY_FUNCTION__);
}

- (void)session:(MCSession *)session didFinishReceivingResourceWithName:(NSString *)resourceName
	   fromPeer:(MCPeerID *)peerID atURL:(NSURL *)localURL withError:(NSError *)error {
	DBLog(@"This app is not configured to receive data with method %s", __PRETTY_FUNCTION__);
}

- (void)session:(MCSession *)session didReceiveStream:(NSInputStream *)stream
	   withName:(NSString *)streamName fromPeer:(MCPeerID *)peerID {
	DBLog(@"This app is not configured to receive data with method %s", __PRETTY_FUNCTION__);
}

- (void)session:(MCSession *)session didReceiveData:(NSData *)data fromPeer:(MCPeerID *)peerID {
	[self receivedData:data fromPeer:peerID.displayName];
}

- (void)session:(MCSession *)session peer:(MCPeerID *)peerID didChangeState:(MCSessionState)state {
	NSString* deviceUID = [self deviceUIDForPeerID:peerID.displayName];

	switch (state) {
		case MCSessionStateConnected:
			DBLog(@"Connected to peer %@", peerID.displayName);
			[self.connectedPeerObjects setObject:peerID forKey:peerID.displayName];
			
			[self sendObject:self.delegate.thisDevice.uid
				 withCommand:SyncDatabaseSyncIdentificationCommand
					 toPeers:@[peerID.displayName]

				wantsReceipt:NO];
			break;
		case MCSessionStateNotConnected:
			[self.discoveredPeers removeObject:peerID.displayName];
			[self.connectedPeerObjects removeObjectForKey:peerID.displayName];

			[self removePeerIdentifier:peerID.displayName];
			[self receivedDisconnection:deviceUID fromPeer:peerID.displayName];
			DBLog(@"Disconnected from peer %@", peerID.displayName);
			break;
		
		case MCSessionStateConnecting:
			DBLog(@"Connecting to peer %@", peerID.displayName);
			break;
		default:
			DBLog(@"Peer %@ has unknown state", peerID.displayName);
			break;
	}
}

#pragma mark - Overridden methods

- (void)receivedIdentification:(NSString *)deviceUID fromPeer:(NSString *)peer {
	[self addConnectedDevice:deviceUID withPeerIdentifier:peer];
	[super receivedIdentification:deviceUID fromPeer:peer];
}

@end
